/* gnubiff -- a mail notification program
 * Copyright (C) 2000-2003 Nicolas Rougier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this software; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */
#include <map>
#include <sstream>
#include <sys/stat.h>
#include <utime.h>
#include "Biff.h"
#include "AppletGTK.h"
#ifdef USE_GNOME
#  include "AppletGNOME.h"
#endif
#include "Mailbox.h"
#include "File.h"
#include "Mh.h"
#include "Maildir.h"
#include "Imap4.h"
#include "Pop3.h"
#include "Apop.h"
#include "Popup.h"



// ===================================================================
// = Callbacks "C" binding ===========================================
// ===================================================================
extern "C" {
  /** "C" binding for the XML parser (see glib doc) */
  void BIFF_xml_start_element (GMarkupParseContext *context, const gchar *element_name,const gchar **attribute_names, const gchar **attribute_values, gpointer data, GError **error) 
  {((Biff *) data)->xml_start_element (context, element_name, attribute_names, attribute_values, error);};
  /** "C" binding for the XML parser (see glib doc) */
  void BIFF_xml_error (GMarkupParseContext *context, GError *error, gpointer data)
  {((Biff *) error)->xml_error (context, error);};
}

// ===================================================================
// = Static attributes ===============================================
// ===================================================================


// ===================================================================
// = Biff ===========================================================
/**
   Create a new Biff object by setting default parameters, then
   loading the configuration file from filename if it exists. If the
   given configuration file is empty, then a default one
   ($HOME/.gnubiffrc) is created from default parameters.
   @param mode is the mode to use (GNOME/GTK) for GNOME version
   @param filename is the configuration file to use
*/
// ===================================================================
Biff::Biff (gint mode, string filename) {
  // Create the crypt table for encryption and decryption of password
  // This table is dependent on the "/bin" directory change time It
  // should be sufficiently different from one system to the other so
  // even if someone stoled the file, he won't be able to decrypt it
  // on his system without this information (which is easily
  // accessible anyway).
  // Note that this time should be pretty stable over time.
  _table = "";

  // Currently disabled
  /*
  struct stat file_stat;
  time_t t;
  if (stat ("/bin", &file_stat) == 0) {
    t = file_stat.st_ctime;
    string d = ctime (&t);
    for (guint i=0; i<d.size(); i++)
      if ((d[i] != '\n') &&  (d[i] != ' ') && (d[i] != ':') && (_table.find(d[i]) == string::npos))
	_table += d[i];
  }
  */
  // In case we don't have enough characters
  _table += "/!&([{)]}+=-*?!$";

  // Interface
  _mode = mode;

  // Setup
  _setup = new Setup (this, GNUBIFF_DATADIR"/Setup.glade2");
  _setup->create ();

  // Applet
#ifdef USE_GNOME
  if (_mode == GNOME_MODE)
    _applet = new AppletGNOME (this, GNUBIFF_DATADIR"/AppletGNOME.glade2");
  else
    _applet = new AppletGTK (this, GNUBIFF_DATADIR"/AppletGTK.glade2");
#else
  _applet = new AppletGTK (this, GNUBIFF_DATADIR"/AppletGTK.glade2");
#endif
  _applet->create();

  // Popup
  _popup = new Popup (this,  GNUBIFF_DATADIR"/Popup.glade2");
  _popup->create();

  // Mailbox
  _mailbox = new File (this);

  // Set default values
  _title = "Mail";
  _protocol = PROTOCOL_FILE;
  _port = 0;
#ifdef HAVELIBSSL
  _ssl = true;
#else
  _ssl = false;
#endif
  _address = "";
  if (g_getenv("MAIL"))
    _address = g_getenv ("MAIL");
  _user = "";
  if (g_get_user_name())
    _user = g_get_user_name();
  _password = "";
  _certificate = "";
  _polltime = 60;
  _poptime = 5;
  _sound_type = SOUND_FILE;
  _sound_file = REAL_DATADIR"/sounds/gnubiff/coin.wav";
  _sound_volume = 100;
  _state = STATE_RUNNING;
  _display_popup = true;
  _mail_app = "xemacs";
#  ifdef USE_GNOME
  if (_mode == USE_GNOME) {
    _mail_image = GNUBIFF_DATADIR"/gnome-mail.gif";
    _nomail_image = GNUBIFF_DATADIR"/gnome-nomail.png";
  }
  else {
    _mail_image = GNUBIFF_DATADIR"/gtk-mail.gif";
    _nomail_image = GNUBIFF_DATADIR"/gtk-nomail.png";
  }
#  else
  _mail_image = GNUBIFF_DATADIR"/gtk-mail.gif";
  _nomail_image = GNUBIFF_DATADIR"/gtk-nomail.png";
#  endif
  _hide_nomail = FALSE;
  _display_title = true;
  _display_date = true;
  _max_sender_size = 50;
  _max_subject_size = 50;
  _popup_x_sign = SIGN_MINUS;
  _popup_x = 0;
  _popup_y_sign = SIGN_PLUS;
  _popup_y = 0;
  _popup_decorated = FALSE;
  _popup_no_geometry = FALSE;
  _biff_x_sign = SIGN_PLUS;
  _biff_x = 0;
  _biff_y_sign = SIGN_PLUS;
  _biff_y = 0;
  _biff_decorated = FALSE;
  _biff_no_geometry = FALSE;
  _max_header_display = 50;
  _popup_title_font = "Sans 10"; 
  _popup_body_font = "Sans 10";
  _biff_title_font = "Sans 8";
  _biff_body_font = "Sans 8";
  _applet_title_font = "Sans 8";
  _applet_body_font = "Sans 8";
  _popup_title_color = "Black"; 
  _popup_body_color = "Blue";
  _biff_title_color = "Black";
  _biff_body_color = "Blue";
  _applet_title_color = "Black";
  _applet_body_color = "Blue";

  // Do we have a valid filename ?
  if (!filename.empty()) {
    _filename = filename;
    load();
  }
  // no ? => use default one and try to load it (if it exists)
  else {
    _filename = string (g_get_home_dir ()) + string ("/.gnubiffrc");
    // Does this file exist ?
    ifstream file;
    file.open (_filename.c_str());
    if (file.is_open()) {
      file.close();
      load();
    }
    else
      file.close();
  }

  // Update GUI
  _setup->update();
}


// ===================================================================
// = ~Biff ==========================================================
// ===================================================================
Biff::~Biff (void) {
}


// ===================================================================
// = save ============================================================
/**
   This function tries to save the current biff into _filename.
   @return false if save failed, true else
*/
// ===================================================================
gboolean Biff::save (void) {
  // Open file
  ofstream file;
  file.open (_filename.c_str());
  if (!file.is_open())
    return false;
  
  // Header
  file << "<?xml version=\"1.0\"?>" << endl;
  file << "<configuration-file>"<< endl;


  // Mailbox
  file << "  <mailbox>" << endl;
  file << "    <parameter name=\"title\"            value=\"" << _title << "\"/>" << endl;
  file << "    <parameter name=\"protocol\"         value=\"" << _protocol << "\"/>" << endl;
  file << "    <parameter name=\"port\"             value=\"" << _port << "\"/>" << endl;
  file << "    <parameter name=\"ssl\"              value=\"" << _ssl << "\"/>" << endl;
  file << "    <parameter name=\"address\"          value=\"" << _address << "\"/>" << endl;
  file << "    <parameter name=\"user\"             value=\"" << _user << "\"/>" << endl;

  // Here we are: we save password.
  // pop3 and imap4 protocols requires password in clear so we have to save password
  // in clear within configuration file. No need to say this is higly unsecure if
  // somebody looks at the file.
  // So we have to take some measures:
  //   1. The file is made readable by owner only.
  //   2. password is "crypted" so it's not directly human readable.
  // Of course, these measures won't prevent a determined person to break in but will
  // at least prevent "ordinary" people to not be able to steal password easily.
#ifdef USE_PASSWORD
  file << "    <parameter name=\"password\"         value=\"";
  for (guint i=0; i<_password.size(); i++)
    file << _table[_password[i]/16] << _table[_password[i]%16];
  file << "\"/>" << endl;
#else
  file << "    <parameter name=\"password\"         value=\"" << "" << "\"/>" << endl;
#endif

  file << "    <parameter name=\"certificate\"      value=\"" << _certificate << "\"/>" << endl;
  file << "  </mailbox>" << endl;


  // Options
  file << "  <options>" << endl;
  file << "    <parameter name=\"polltime\"         value=\"" << _polltime << "\"/>" << endl;
  file << "    <parameter name=\"poptime\"          value=\"" << _poptime << "\"/>" << endl;
  file << "    <parameter name=\"sound_type\"       value=\"" << _sound_type << "\"/>" << endl;
  file << "    <parameter name=\"sound_file\"       value=\"" << _sound_file << "\"/>" << endl;
  file << "    <parameter name=\"sound_volume\"     value=\"" << _sound_volume << "\"/>" << endl;
  file << "    <parameter name=\"state\"            value=\"" << _state << "\"/>" << endl;

  
  //  => Advanced
  file << "    <advanced>" << endl;
  file << "      <parameter name=\"display_popup\"      value=\"" << _display_popup << "\"/>" << endl;
  file << "      <parameter name=\"mail_app\"           value=\"" << _mail_app << "\"/>" << endl;
  file << "    </advanced>" << endl;

  // End Options
  file << "  </options>" << endl;


  // Layout
  file << "  <layout>" << endl;
  file << "    <parameter name=\"mail_image\"       value=\"" << _mail_image << "\"/>" << endl;
  file << "    <parameter name=\"nomail_image\"     value=\"" << _nomail_image << "\"/>" << endl;
  file << "    <parameter name=\"hide_nomail\"      value=\"" << _hide_nomail << "\"/>" << endl;
  file << "    <parameter name=\"display_title\"    value=\"" << _display_title << "\"/>" << endl;
  file << "    <parameter name=\"display_date\"     value=\"" << _display_date << "\"/>" << endl;
  file << "    <parameter name=\"max_sender_size\"  value=\"" << _max_sender_size << "\"/>" << endl;
  file << "    <parameter name=\"max_subject_size\" value=\"" << _max_subject_size << "\"/>" << endl;

  
  //  => Geometry
  file << "    <geometry>" << endl;
  file << "      <parameter name=\"popup_x_sign\"       value=\"" << _popup_x_sign << "\"/>" << endl;
  file << "      <parameter name=\"popup_x\"            value=\"" << _popup_x << "\"/>" << endl;
  file << "      <parameter name=\"popup_y_sign\"       value=\"" << _popup_y_sign << "\"/>" << endl;
  file << "      <parameter name=\"popup_y\"            value=\"" << _popup_y << "\"/>" << endl;
  file << "      <parameter name=\"popup_decorated\"    value=\"" << _popup_decorated << "\"/>" << endl;
  file << "      <parameter name=\"popup_no_geometry\"  value=\"" << _popup_no_geometry << "\"/>" << endl;
  file << "      <parameter name=\"biff_x_sign\"        value=\"" << _biff_x_sign << "\"/>" << endl;
  file << "      <parameter name=\"biff_x\"             value=\"" << _biff_x << "\"/>" << endl;
  file << "      <parameter name=\"biff_y_sign\"        value=\"" << _biff_y_sign << "\"/>" << endl;
  file << "      <parameter name=\"biff_y\"             value=\"" << _biff_y << "\"/>" << endl;
  file << "      <parameter name=\"biff_decorated\"     value=\"" << _biff_decorated << "\"/>" << endl;
  file << "      <parameter name=\"biff_no_geometry\"   value=\"" << _biff_no_geometry << "\"/>" << endl;
  file << "      <parameter name=\"max_header_display\" value=\"" << _max_header_display << "\"/>" << endl;
  file << "    </geometry>" << endl;


  //  => Font
  file << "    <font>" << endl;
  file << "      <parameter name=\"popup_title_font\"   value=\"" << _popup_title_font << "\"/>" << endl;
  file << "      <parameter name=\"popup_body_font\"    value=\"" << _popup_body_font << "\"/>" << endl;
  file << "      <parameter name=\"biff_title_font\"    value=\"" << _biff_title_font << "\"/>" << endl;
  file << "      <parameter name=\"biff_body_font\"     value=\"" << _biff_body_font << "\"/>" << endl;
  file << "      <parameter name=\"applet_title_font\"  value=\"" << _applet_title_font << "\"/>" << endl;
  file << "      <parameter name=\"applet_body_font\"   value=\"" << _applet_body_font << "\"/>" << endl;
  file << "      <parameter name=\"popup_title_color\"   value=\"" << _popup_title_color << "\"/>" << endl;
  file << "      <parameter name=\"popup_body_color\"    value=\"" << _popup_body_color << "\"/>" << endl;
  file << "      <parameter name=\"biff_title_color\"    value=\"" << _biff_title_color << "\"/>" << endl;
  file << "      <parameter name=\"biff_body_color\"     value=\"" << _biff_body_color << "\"/>" << endl;
  file << "      <parameter name=\"applet_title_color\"  value=\"" << _applet_title_color << "\"/>" << endl;
  file << "      <parameter name=\"applet_body_color\"   value=\"" << _applet_body_color << "\"/>" << endl;
  file << "    </font>" << endl;


  // End Layout
  file << "  </layout>" << endl;

  // End Header
  file << "</configuration-file>"<< endl;

  // Close the file
  file.close ();

  // Restrict file permission
  string command = string("chmod 600 ") + _filename;
  system (command.c_str());
  
  return true;
}


// ===================================================================
// = load ============================================================
/**
   This function tries to load a biff configuration file. The XML
   parsing is done via glib tiny xml parser.
   @return false if save failed, true else
*/
// ===================================================================
gboolean Biff::load (void) {
  // Open the file
  ifstream file;
  string line;
  file.open (_filename.c_str());
  if (!file.is_open())
    return false;

  // Instantiate a new xml parser
  GMarkupParser parser;
  parser.start_element = BIFF_xml_start_element;
  parser.end_element = 0;
  parser.text = 0;
  parser.passthrough = 0;
  parser.error = BIFF_xml_error;
  GMarkupParseContext *context = g_markup_parse_context_new (&parser, GMarkupParseFlags (0), this, 0);

  // Parse the file
  gboolean status = TRUE;
  while ((!getline(file, line).eof()) && (status))
    status = g_markup_parse_context_parse (context, line.c_str(), line.size(), 0);

  // Instantiate protocol
  delete _mailbox;
  switch (_protocol) {
  case PROTOCOL_FILE:
    _mailbox = new File (this);
    break;
  case PROTOCOL_MH:
    _mailbox = new Mh (this);
    break;
  case PROTOCOL_MAILDIR:
    _mailbox = new Maildir (this);
    break;
  case PROTOCOL_IMAP4:
    _mailbox = new Imap4 (this);
    break;
  case PROTOCOL_POP3:
    _mailbox = new Pop3 (this);
    break;
  case PROTOCOL_APOP:
    _mailbox = new Apop (this);
    break;
  }

  // Close the file
  file.close ();
  return true;
}

// ===================================================================
// = xml_start_element ===============================================
/**
   This function is called by g_markup_parse_context_parse at the
   start of a new xml element. See glib documentation for further
   information.
*/
// ===================================================================
void Biff::xml_start_element (GMarkupParseContext *context,
			      const gchar *element_name,
			      const gchar **attribute_names,
			      const gchar **attribute_values,
			      GError **error) {
  if (string (element_name) == "parameter") {
    // Store attributes (name & value) in a map
    map<string,string> fmap;
    guint i = 0;
    while (attribute_names[i]) {
      fmap[attribute_names[i]] = attribute_values[i];
      i++;
    }
    


    // Mailbox
    if (fmap["name"] == "title")
      _title = fmap["value"];
    else if (fmap["name"] == "protocol") {
      istringstream strin(fmap["value"]);
      strin >> _protocol;
    }
    else if (fmap["name"] == "port") {
      istringstream strin(fmap["value"]);
      strin >> _port;
    }
    else if (fmap["name"] == "ssl") {
      istringstream strin(fmap["value"]);
      strin >> _ssl;
    }
    else if (fmap["name"] == "address")
      _address = fmap["value"];
    else if (fmap["name"] == "user")
      _user = fmap["value"];
    else if (fmap["name"] == "password") {
      _password = "";
      string tmp = fmap["value"];
      for (gint i=0; i<gint(tmp.size())-1; i+=2) {
	char c = 0;
	guint j;
	for (j=0; j<16; j++)
	  if (_table [j] == tmp[i])
	    c += j*16;
	  else if (_table [j] == tmp[i+1])
	    c += j;
	_password += c;
      }
    }
    else if (fmap["name"] == "certificate")
      _certificate = fmap["value"];


    // Options
    else if (fmap["name"] == "polltime") {
      istringstream strin(fmap["value"]);
      strin >> _polltime;
    }
    else if (fmap["name"] == "poptime") {
      istringstream strin(fmap["value"]);
      strin >> _poptime;
    }
    else if (fmap["name"] == "sound_type") {
      istringstream strin(fmap["value"]);
      strin >> _sound_type;
    }
    else if (fmap["name"] == "sound_volume") {
      istringstream strin(fmap["value"]);
      strin >> _sound_volume;
    }
    else if (fmap["name"] == "sound_file") {
      _sound_file = fmap["value"];
      if (!g_file_test (_sound_file.c_str(), G_FILE_TEST_EXISTS))
	_sound_file = REAL_DATADIR"/sounds/gnubiff/coin.wav"; 
    }
    else if (fmap["name"] == "state") {
      istringstream strin(fmap["value"]);
      strin >> _state;
    }



    // Advanced
    else if (fmap["name"] == "display_popup") {
      istringstream strin(fmap["value"]);
      strin >> _display_popup;
    }
    else if (fmap["name"] == "mail_app")
      _mail_app = fmap["value"];



    // Layout
    else if (fmap["name"] == "mail_image") {
      _mail_image = fmap["value"];
      if (!g_file_test (_mail_image.c_str(), G_FILE_TEST_EXISTS)) {
#       ifdef USE_GNOME
	if (_mode == GNOME_MODE)
	  _mail_image = GNUBIFF_DATADIR"/gnome-mail.gif";
	else
	  _mail_image = GNUBIFF_DATADIR"/gtk-mail.gif";
#       else
	_mail_image = GNUBIFF_DATADIR"/gtk-mail.gif";
#       endif
      }
    }
    else if (fmap["name"] == "nomail_image") {
      _nomail_image = fmap["value"];
      if (!g_file_test (_nomail_image.c_str(), G_FILE_TEST_EXISTS)) {
#       ifdef USE_GNOME
	if (_mode == GNOME_MODE)
	  _nomail_image = GNUBIFF_DATADIR"/gnome-nomail.png";
	else
	  _nomail_image = GNUBIFF_DATADIR"/gtk-nomail.png";
#        else
	_nomail_image = GNUBIFF_DATADIR"/gtk-nomail.png";
#       endif
      }
    }
    else if (fmap["name"] == "hide_nomail") {
      istringstream strin(fmap["value"]);
      strin >> _hide_nomail;
    }
    else if (fmap["name"] == "display_title") {
      istringstream strin(fmap["value"]);
      strin >> _display_title;
    }
    else if (fmap["name"] == "display_date") {
      istringstream strin(fmap["value"]);
      strin >> _display_date;
    }
    else if (fmap["name"] == "max_sender_size") {
      istringstream strin(fmap["value"]);
      strin >> _max_sender_size;
    }
    else if (fmap["name"] == "max_subject_size") {
      istringstream strin(fmap["value"]);
      strin >> _max_subject_size;
    }



    // Geometry
    else if (fmap["name"] == "popup_x_sign") {
      istringstream strin(fmap["value"]);
      strin >> _popup_x_sign;
    }
    else if (fmap["name"] == "popup_x") {
      istringstream strin(fmap["value"]);
      strin >> _popup_x;
    }
    else if (fmap["name"] == "popup_y_sign") {
      istringstream strin(fmap["value"]);
      strin >> _popup_y_sign;
    }
    else if (fmap["name"] == "popup_y") {
      istringstream strin(fmap["value"]);
      strin >> _popup_y;
    }
    else if (fmap["name"] == "popup_decorated") {
      istringstream strin(fmap["value"]);
      strin >> _popup_decorated;
    }
    else if (fmap["name"] == "popup_no_geometry") {
      istringstream strin(fmap["value"]);
      strin >> _popup_no_geometry;
    }
    else if (fmap["name"] == "biff_x_sign") {
      istringstream strin(fmap["value"]);
      strin >> _biff_x_sign;
    }
    else if (fmap["name"] == "biff_x") {
      istringstream strin(fmap["value"]);
      strin >> _biff_x;
    }
    else if (fmap["name"] == "biff_y_sign") {
      istringstream strin(fmap["value"]);
      strin >> _biff_y_sign;
    }
    else if (fmap["name"] == "biff_y") {
      istringstream strin(fmap["value"]);
      strin >> _biff_y;
    }
    else if (fmap["name"] == "biff_decorated") {
      istringstream strin(fmap["value"]);
      strin >> _biff_decorated;
    }
    else if (fmap["name"] == "biff_no_geometry") {
      istringstream strin(fmap["value"]);
      strin >> _biff_no_geometry;
    }
    else if (fmap["name"] == "max_header_display") {
      istringstream strin(fmap["value"]);
      strin >> _max_header_display;
    }



    // Font
    else if (fmap["name"] == "popup_title_font")
      _popup_title_font = fmap["value"];
    else if (fmap["name"] == "popup_body_font")
      _popup_body_font = fmap["value"];
    else if (fmap["name"] == "biff_title_font")
      _biff_title_font = fmap["value"];
    else if (fmap["name"] == "biff_body_font")
      _biff_body_font = fmap["value"];
    else if (fmap["name"] == "applet_title_font")
      _applet_title_font = fmap["value"];
    else if (fmap["name"] == "applet_body_font")
      _applet_body_font = fmap["value"];
    else if (fmap["name"] == "popup_title_color")
      _popup_title_color = fmap["value"];
    else if (fmap["name"] == "popup_body_color")
      _popup_body_color = fmap["value"];
    else if (fmap["name"] == "biff_title_color")
      _biff_title_color = fmap["value"];
    else if (fmap["name"] == "biff_body_color")
      _biff_body_color = fmap["value"];
    else if (fmap["name"] == "applet_title_color")
      _applet_title_color = fmap["value"];
    else if (fmap["name"] == "applet_body_color")
      _applet_body_color = fmap["value"];
  }
}


// ===================================================================
// = xml_error =======================================================
/**
   This function is called by g_markup_parse_context_parse if error
   occured while parsing. See glib documentation for further
   information.
*/
// ===================================================================
void Biff::xml_error (GMarkupParseContext *context, GError *error) {
  g_warning ("%s\n", error->message);
}

